#!/bin/sh
#
# Shell script to implent HTML cross-reference checking with htmlchek.pl
#
if test `pwd` != $HOME/public_html
then echo "Current directory is not \$HOME/public_html (this may not be what you intended)."
fi
case ${HTMLCHEK:-"/"} in */ )
if test $# -lt 2
then
    echo "
 Syntax is:  $0 directoryprefix outfileprefix [optional params]

$0 should be run with the current directory set to the directory"
cat << '_EOHERE_'
which is at the root of the tree of HTML files to be examined (for example,
"$HOME/public_html").  The first parameter ``directoryprefix'' (which
should end with a slash `/') is how this top directory is referenced in
<... HREF="..."> links in files in subordinate directories in the tree (for
example, as "/~myself/" or "http://myhost.edu/~myself/").  If the files
you are checking refer to each other only with simple relative URL's
(i.e. with no access method or hostname specified, and not beginning with
`/'), then you can get away with specifying ``directoryprefix'' as a null
string ('').

  The second parameter ``outfileprefix'' is the name of the files (with the
extensions ".NAME", ".HREF", ".SRC", and ".ERR") in which the ouput of the
HTML-checking and cross-referencing process will be put.  And finally, the
optional parameters are all of the form "option=value", as explained in the
_EOHERE_
echo "docs.  So the following are some typical command lines (putting $0
first in a command, as in the first example, implies that you have set execute
permission by running \"chmod +x $0\"):

  $0  http://uts.cc.utexas.edu/~churchh/  check

  cd  \$HOME/public_html  ;  sh  $0  ''  out  html3=1
"
    exit 1
else
    if test -s ${HTMLCHEK}htmlchek.pl
    then
    case ${1:-"/"} in */ )
     case $1
     in *?=* ) echo "Error: dirprefix parameter  $1  is in option=value form"
               exit 1;;
     esac
     case $2
     in *?=* ) echo "Error: outfilesname parameter  $2  is in option=value form"
               exit 1;;
     esac
    oprefix=$2
    dprefix=$1
    shift 2
    for px
    do
    case $px
    in -?* ) echo "You are passing flags to the perl interpreter in argument $px"
             echo "(This may not be what you intended)";;
    append=* | arena=* | configfile=* | deprecated=* | html3=* | htmlplus=* |\
    loosepair=* | lowlevelnonpair=* | lowlevelpair=* | netscape=* |\
    nonblock=* | nonpair=* | nonrecurpair=* | nowswarn=* |\
    reqopts=* | strictpair=* | sugar=* | tagopts=* | usebase=* ) ;;
       * ) echo "Error: The argument  $px  is not in option=value form, where,"
           echo "the part before the equals sign \`=' is a recognized option."
           exit 1;;
    esac
    done
    echo "Checking all .html files in and under `pwd` now... (using perl)"
    perl ${HTMLCHEK}htmlchek.pl dirprefix=$dprefix refsfile=$oprefix $@ `find . -name \*.html -print` > $oprefix.ERR
    if test ! \( -s $oprefix.NAME -a -s $oprefix.HREF -a -s $oprefix.SRC -a -s $oprefix.ERR \)
    then echo "Was some error on output files; Exiting..."
         exit 1
    else echo "Sorting and correlating files now..."
    fi
    sort < $oprefix.NAME | uniq > $oprefix.QNAME
    sort < $oprefix.HREF | uniq > $oprefix.QHREF
    diff $oprefix.QNAME $oprefix.QHREF > $oprefix.TMP
    perl -ne 'if ($_ =~ /^< /) {print substr($_,2)}' $oprefix.TMP > $oprefix.QHREF
    perl -ne 'if ($_ =~ /^> /) {print substr($_,2)}' $oprefix.TMP > $oprefix.QNAME
    diff $oprefix.QNAME $oprefix.QHREF > $oprefix.TMP                 # Iterate
    perl -ne 'if ($_ =~ /^< /) {print substr($_,2)}' $oprefix.TMP > $oprefix.QHREF
    perl -ne 'if ($_ =~ /^> /) {print substr($_,2)}' $oprefix.TMP > $oprefix.QNAME
    sort < $oprefix.SRC | uniq > $oprefix.TMP
    mv $oprefix.TMP $oprefix.SRC      # You can change this if you don't want
    mv $oprefix.QNAME $oprefix.NAME   # to overwrite the original three files
    mv $oprefix.QHREF $oprefix.HREF
    echo "Errorcheck data is in the file $oprefix.ERR; locations in the specified files
 which are not referenced from the files are in $oprefix.NAME, references from
 the specified files which are not found in the files are in $oprefix.HREF, and
 references to inline images are in $oprefix.SRC";;
    * ) echo "
First parameter (dirprefix) does not end in \`/'"
        exit 1;;
    esac
    else echo "
htmlchek.pl is not found.  Either copy it to the current directory, or
set the environment variable HTMLCHEK to the pathname where it is located
(this should terminate with a \`/' character).  Do \`setenv HTMLCHEK /somedir/'
in csh and tcsh, \`HTMLCHEK=/somedir/;export HTMLCHEK' in sh and its offspring."
         exit 1
    fi
fi;;
* ) echo "Environment variable HTMLCHEK does not end in \`/'"
    exit 1;;
esac
exit 0
